#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <stdio.h>
#include <string.h>

#include "mctd.h"
#include "mctd-log.h"

#define SERIAL_TTY "ttymxc1"
#define SERIAL_PORT "/dev/" SERIAL_TTY
#define SERIAL_LOCK_FILE "/var/lock/LCK.." SERIAL_TTY
#define SERIAL_LOCK_PID_FILE DAEMON_PATH "serial_lock.pid"
#define SERIAL_BAUDRATE B460800
#define SERIAL_BAUDRATE_DEFAULT B115200

static bool is_using()
{
    int fd;

    fd = open(SERIAL_LOCK_FILE, O_RDONLY);
    if (fd == -1)
        return false;

    close(fd);
    return true;
}

static int lock()
{
    int fd;
    int fd_pid;
    char buf[11];

    if (is_using())
        return -MCTD_RESULT_ERR_USING_SERIAL;

    fd = open(SERIAL_LOCK_FILE, O_WRONLY | O_CREAT, 0666);
    if (fd == -1) {
        perror("open serial lock file");
        return -MCTD_RESULT_ERR_SERIAL;
    }

    fd_pid = open(SERIAL_LOCK_PID_FILE, O_WRONLY | O_CREAT, 0666);

    sprintf(buf, "%10d", getpid());
    write(fd, buf, 10);

    if (fd_pid != -1) {
        write(fd_pid, buf, 10);
        close(fd_pid);
    } else
        perror("open lock pid file");

    close(fd);
    return MCTD_RESULT_OK;
}

static void unlock()
{
    remove(SERIAL_LOCK_FILE);
    remove(SERIAL_LOCK_PID_FILE);
}

static struct termios tio_orig;
static bool has_tio_orig = false;

static void restore(int fd)
{
    if (has_tio_orig)
        tcsetattr(fd, TCSANOW, &tio_orig);

    has_tio_orig = false;
}

static void setup(int fd, bool default_baudrate)
{
    struct termios tio;
    int ret;

    ret = tcgetattr(fd, &tio_orig);
    if (ret) {
        perror("tcgetattr");
        return;
    }
    has_tio_orig = true;

    memset (&tio, 0, sizeof(struct termios));
    /*tio.c_cflag &= ~(CBAUD | CSIZE | CSTOPB | PARENB | PARODD | CRTSCTS);*/
    tio.c_cflag &= ~(CBAUD | CSIZE | CSTOPB | PARENB | PARODD);
    tio.c_iflag &= ~(IGNCR | ICRNL | IUCLC | INPCK | IXON | IXOFF | IXANY );
    tio.c_oflag &= ~(OPOST | OLCUC | OCRNL | ONLCR | ONLRET);
    tio.c_lflag &= ~(ICANON | ECHO | ECHOE | ECHONL);
    tio.c_cc[VMIN] = 1;
    tio.c_cc[VTIME] = 0;
    tio.c_cc[VEOF] = 1;
    tio.c_iflag |= IGNPAR;
    tio.c_cflag |= (CS8 | CLOCAL | CREAD | CRTSCTS);

    ret = cfsetispeed(&tio, SERIAL_BAUDRATE);
    if (ret < 0) {
        perror("cfsetispeed");
        return;
    }

    ret = cfsetospeed(&tio, default_baudrate ? SERIAL_BAUDRATE_DEFAULT
                                             : SERIAL_BAUDRATE);
    if (ret < 0) {
        perror("cfsetospeed");
        return;
    }

    ret = tcflush(fd, TCIFLUSH);
    if (ret < 0)
        perror("tcflush");

    ret = tcsetattr(fd, TCSANOW, &tio);
    if (ret)
        perror("tcsetattr");
}

static int open_with_default_baudrate(bool default_baudtrate)
{
    int fd;
    int result;

    result = lock();
    if (result != MCTD_RESULT_OK) {
        log_warn("lock failed");
        return result;
    }

    fd = open(SERIAL_PORT, O_RDWR);
    if (fd == -1) {
        perror("open serial");
        log_warn("open failed");
        unlock();
        return MCTD_RESULT_ERR_SERIAL;
    }

    setup(fd, default_baudtrate);
    return fd;
}

int mctd_serial_open_baudrate_115200()
{
    return open_with_default_baudrate(true);
}

int mctd_serial_open()
{
    return open_with_default_baudrate(false);
}

void mctd_serial_close(int fd)
{
    restore(fd);
    unlock();
    close(fd);
}

void mctd_serial_remove_own_lock_file()
{
    int fd;
    int fd_pid;

    char buf[16];
    char buf_pid[16];

    fd = open(SERIAL_LOCK_FILE, O_RDONLY);
    if (fd == -1) {
        return;
    }

    fd_pid = open(SERIAL_LOCK_PID_FILE, O_RDONLY);
    if (fd_pid == -1) {
        close(fd);
        return;
    }

    memset(buf, 0, 16);
    memset(buf_pid, 0, 16);

    read(fd, buf, 16);
    read(fd_pid, buf_pid, 16);

    close(fd);
    close(fd_pid);

    if (strcmp(buf, buf_pid) == 0) {
        log_info("erase lock file");
        unlock();
    }
}

const char *mctd_serial_port_str()
{
    return SERIAL_PORT;
}

int mctd_serial_convert_error(int error)
{
    switch(error) {
    case -MCTD_RESULT_ERR_USING_SERIAL: return MCTD_RESULT_ERR_USING_SERIAL;
    default:                            return MCTD_RESULT_ERR_SERIAL;
    }
}
